# -*- coding: utf-8 -*-
# file: boosting_core.py
# time: 2021/12/20
# author: yangheng <yangheng@m.scnu.edu.cn>
# github: https://github.com/yangheng95
# Copyright (C) 2021. All Rights Reserved.
import copy
import itertools
import random
import shutil

import tqdm

from findfile import find_cwd_files, find_cwd_dir
from pyabsa import APCConfigManager

import os

from findfile import find_files, find_dir
from termcolor import colored

from pyabsa.functional import Trainer
from pyabsa.functional.dataset import DatasetItem
from pyabsa.functional.dataset.dataset_manager import download_datasets_from_github, ABSADatasetList


def boost_free_training(config: APCConfigManager,
                        dataset: DatasetItem,
                        resplit_dataset=True,
                        device='cuda'):
    prepare_dataset_and_clean_env()
    if find_cwd_dir('checkpoints'):
        shutil.rmtree(find_cwd_dir('checkpoints'))
    if resplit_dataset:
        data = []
        for dataset_file in detect_dataset(dataset, 'apc')['train'] + detect_dataset(dataset, 'apc')['test']:
            print('processing {}'.format(dataset_file))
            fin = open(dataset_file, encoding='utf8', mode='r')
            lines = fin.readlines()
            fin.close()
            os.remove(dataset_file)
            for i in tqdm.tqdm(range(0, len(lines), 3)):
                lines[i] = lines[i].strip()
                lines[i + 1] = lines[i + 1].strip()
                lines[i + 2] = lines[i + 2].strip()

                data.append([lines[i], lines[i + 1], lines[i + 2]])

        random.shuffle(data)

        train_data = data[len(data) // 5:]
        boost_data = data[:len(data) // 5]

        fout_train = open('{}/train.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')
        fout_test = open('{}/test.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')
        for case in train_data:
            for line in case:
                fout_train.write(line + '\n')

        for case in boost_data:
            for line in case:
                fout_test.write(line + '\n')

        fout_train.close()
        fout_test.close()

    Trainer(config=config,
            dataset=dataset,  # train set and test set will be automatically detected
            checkpoint_save_mode=0,  # =None to avoid save model
            auto_device=device  # automatic choose CUDA or CPU
            )


def classic_boost_training(config: APCConfigManager,
                           dataset: DatasetItem,
                           n_fold=5,
                           resplit_dataset=True,
                           aug_tool='textattack',
                           device='cuda'):
    prepare_dataset_and_clean_env()
    if find_cwd_dir('checkpoints'):
        shutil.rmtree(find_cwd_dir('checkpoints'))
    if aug_tool in 'textattack':
        from textattack.augmentation import EasyDataAugmenter as Aug
        # Alter default values if desired
        augmenter = Aug(pct_words_to_swap=0.3, transformations_per_example=10)
    else:
        import nlpaug.augmenter.word as naw
        augmenter = naw.ContextualWordEmbsAug(
            # model_path='microsoft/deberta-v3-base', action="insert")
            model_path='roberta-base', action="substitute", device=device)

    if resplit_dataset:
        train_data = []
        for dataset_file in detect_dataset(dataset, 'apc')['train'] + detect_dataset(dataset, 'apc')['test']:
            print('processing {}'.format(dataset_file))
            fin = open(dataset_file, encoding='utf8', mode='r')
            lines = fin.readlines()
            fin.close()
            os.remove(dataset_file)
            for i in tqdm.tqdm(range(0, len(lines), 3)):
                lines[i] = lines[i].strip()
                lines[i + 1] = lines[i + 1].strip()
                lines[i + 2] = lines[i + 2].strip()

                train_data.append([lines[i], lines[i + 1], lines[i + 2]])

        random.shuffle(train_data)

        train_data = train_data[len(train_data) // n_fold:]
        test_data = train_data[:len(train_data) // n_fold]

        fout_test = open('{}/test.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')

        for case in test_data:
            for line in case:
                fout_test.write(line + '\n')

        fout_train = open('{}/classic_aug.train.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')

        for item in tqdm.tqdm(train_data, postfix='Augmenting...'):
            fout_train.write(item[0] + '\n')
            fout_train.write(item[1] + '\n')
            fout_train.write(item[2] + '\n')
            item[0] = item[0].replace('$T$', '%%')
            if aug_tool in 'textattack':
                augs = augmenter.augment(item[0])
            else:
                augs = augmenter.augment(item[0], n=10, num_thread=os.cpu_count())
            for aug in augs:
                if '%%' in item[0]:
                    _text = item[0].replace('%%', '$T$')
                    fout_train.write(aug + '\n')
                    fout_train.write(item[1] + '\n')
                    fout_train.write(item[2] + '\n')

        fout_train.close()
        fout_test.close()

    else:
        train_data = []
        for dataset_file in detect_dataset(dataset, 'apc')['train']:
            print('processing {}'.format(dataset_file))
            fin = open(dataset_file, encoding='utf8', mode='r')
            lines = fin.readlines()
            fin.close()
            os.remove(dataset_file)
            for i in tqdm.tqdm(range(0, len(lines), 3)):
                lines[i] = lines[i].strip()
                lines[i + 1] = lines[i + 1].strip()
                lines[i + 2] = lines[i + 2].strip()

                train_data.append([lines[i], lines[i + 1], lines[i + 2]])

        fout_train = open('{}/classic_aug.train.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')

        for item in tqdm.tqdm(train_data, postfix='Augmenting...'):
            fout_train.write(item[0] + '\n')
            fout_train.write(item[1] + '\n')
            fout_train.write(item[2] + '\n')
            item[0] = item[0].replace('$T$', '%%')
            if aug_tool in 'textattack':
                augs = augmenter.augment(item[0])
            else:
                augs = augmenter.augment(item[0], n=10, num_thread=os.cpu_count())
            for aug in augs:
                if '%%' in item[0]:
                    _text = item[0].replace('%%', '$T$')
                    fout_train.write(aug + '\n')
                    fout_train.write(item[1] + '\n')
                    fout_train.write(item[2] + '\n')

        fout_train.close()

    print(colored('Start classic boosting augment...', 'red'))

    config.log_step = len(train_data) // 10
    Trainer(config=config,
            dataset=dataset,  # train set and test set will be automatically detected
            checkpoint_save_mode=0,  # =None to avoid save model
            auto_device=device  # automatic choose CUDA or CPU
            ).load_trained_model()


def cross_feature_boost_training(config: APCConfigManager,
                                 dataset: DatasetItem,
                                 n_fold=5,
                                 resplit_dataset=True,
                                 aug_tool='textattack',
                                 device='cuda'):
    if aug_tool in 'textattack':
        from textattack.augmentation import EasyDataAugmenter as Aug
        # Alter default values if desired
        augmenter = Aug(pct_words_to_swap=0.3, transformations_per_example=10)
    else:
        import nlpaug.augmenter.word as naw
        augmenter = naw.ContextualWordEmbsAug(
            # model_path='microsoft/deberta-v3-base', action="insert")
            model_path='roberta-base', action="substitute", device=device)

    prepare_dataset_and_clean_env()
    data = []
    dataset_file = ''
    if resplit_dataset:
        dataset_files = detect_dataset(dataset, 'apc')['train'] + detect_dataset(dataset, 'apc')['test']

    else:
        dataset_files = detect_dataset(dataset, 'apc')['train']

    for dataset_file in dataset_files:
        print('processing {}'.format(dataset_file))
        fin = open(dataset_file, encoding='utf8', mode='r')
        lines = fin.readlines()
        fin.close()
        os.remove(dataset_file)
        for i in tqdm.tqdm(range(0, len(lines), 3)):
            lines[i] = lines[i].strip()
            lines[i + 1] = lines[i + 1].strip()
            lines[i + 2] = lines[i + 2].strip()

            data.append([lines[i], lines[i + 1], lines[i + 2]])

    if resplit_dataset:

        random.shuffle(data)

        fout_test = open('{}/test.dat'.format(os.path.dirname(dataset_file)), encoding='utf8', mode='w')
        test_data = data[:len(data) // n_fold]
        for case in test_data:
            for line in case:
                fout_test.write(line + '\n')
        fout_test.close()

        data = data[len(data) // 4:]
        len_per_fold = len(data) // 4 + 1
        folds = [data[i: i + len_per_fold] for i in range(0, len(data), len_per_fold)]

    else:
        len_per_fold = len(data) // 4 + 1
        folds = [data[i: i + len_per_fold] for i in range(0, len(data), len_per_fold)]

    for fold_id, b_idx in enumerate(range(len(folds))):
        print(colored('No.{} fold boosting...'.format(b_idx + 1), 'red'))
        if find_cwd_dir('checkpoints'):
            shutil.rmtree(find_cwd_dir('checkpoints'))
        train_data = list(itertools.chain(*[x for i, x in enumerate(folds) if i != b_idx]))
        valid_data = folds[b_idx]

        for train_file in find_cwd_files('train.dat'):
            os.remove(train_file)
        for valid_file in find_cwd_files('valid.dat', exclude_key='.ignore'):
            os.remove(valid_file)

        fout_train = open('{}/train.dat'.format(os.path.dirname(dataset_file), fold_id), encoding='utf8', mode='w')
        fout_boost = open('{}/valid.dat'.format(os.path.dirname(dataset_file), fold_id), encoding='utf8', mode='w')
        for case in train_data:
            for line in case:
                fout_train.write(line + '\n')

        for case in valid_data:
            for line in case:
                fout_boost.write(line + '\n')

        fout_train.close()
        fout_boost.close()

        _config = copy.deepcopy(config)
        _config.seed = [random.randint(0, 10000)]
        _config.log_step = -1
        sent_classifier = Trainer(config=_config,
                                  dataset=dataset,  # train set and test set will be automatically detected
                                  checkpoint_save_mode=0,  # =None to avoid save model
                                  auto_device=device  # automatic choose CUDA or CPU
                                  ).load_trained_model()

        sent_classifier.opt.eval_batch_size = 128
        dataset_files = detect_dataset(dataset, 'apc')
        boost_sets = dataset_files['valid']
        augmentations = []
        origin_data = []

        for boost_set in boost_sets:
            print('Augmenting -> {}'.format(boost_set))
            fin = open(boost_set, encoding='utf8', mode='r')
            lines = fin.readlines()
            fin.close()
            os.remove(boost_set)
            for i in tqdm.tqdm(range(0, len(lines), 3), postfix='Augmenting...'):

                lines[i] = lines[i].strip().replace('$T$', '%%')
                lines[i + 1] = lines[i + 1].strip()
                lines[i + 2] = lines[i + 2].strip()
                if aug_tool in 'textattack':
                    augs = augmenter.augment(lines[i])
                else:
                    augs = augmenter.augment(lines[i], n=10, num_thread=os.cpu_count())

                for text in augs:
                    if '%%' in text:
                        _text = text.replace('%%', '[ASP]{}[ASP] '.format(lines[i + 1])) + ' !sent! {}'.format(lines[i + 2])
                    else:
                        continue
                    results = sent_classifier.infer(_text, print_result=False)
                    # if results[0]['ref_check'][0] == 'Correct' and 0.95 > results[0]['confidence'][0] > 0.9:
                    if results[0]['ref_check'][0] == 'Correct' and results[0]['confidence'][0] > 0.9:

                        augmentations.extend(
                            [text.replace('%%', '$T$').replace('$*$', '$T$').replace('$-$', '$T$').replace('$ * $', '$T$'), lines[i + 1], lines[i + 2]]
                        )
                origin_data.extend(
                    [lines[i].replace('%%', '$T$').replace('$*$', '$T$').replace('$-$', '$T$').replace('$ * $', '$T$'), lines[i + 1], lines[i + 2]]
                )

            fout = open('{}/{}.train.augment.ignore'.format(os.path.dirname(dataset_file), fold_id), encoding='utf8', mode='w')

            for line in origin_data + augmentations:
                fout.write(line + '\n')
            fout.close()

        del sent_classifier

    for f in find_cwd_files('.ignore'):
        os.rename(f, f.replace('.ignore', ''))

    for f in find_cwd_files('train.dat'):
        os.remove(f)

    print(colored('Start cross boosting augment...', 'red'))
    _config = copy.deepcopy(config)
    _config.log_step = len(origin_data + augmentations) // 10
    Trainer(config=_config,
            dataset=dataset,  # train set and test set will be automatically detected
            checkpoint_save_mode=0,  # =None to avoid save model
            auto_device=device  # automatic choose CUDA or CPU
            )


def prepare_dataset_and_clean_env():
    download_datasets_from_github('../')
    if find_cwd_dir('integrated_datasets', disable_alert=True):
        shutil.rmtree(find_cwd_dir('integrated_datasets', disable_alert=True))
    backup_datasets_dir = find_dir('../', 'integrated_datasets', disable_alert=True)
    integrated_datasets_dir = backup_datasets_dir.replace('../', '')
    shutil.copytree(backup_datasets_dir, integrated_datasets_dir)
    if find_cwd_dir('checkpoints'):
        shutil.rmtree(find_cwd_dir('checkpoints'))


def detect_dataset(dataset_path, task='apc'):
    filter_key_words = ['.py', '.ignore', '.md', 'readme', 'log', 'result', 'zip', '.state_dict', '.model', '.png']

    if not isinstance(dataset_path, DatasetItem):
        dataset_path = DatasetItem(dataset_path)
    dataset_file = {'train': [], 'test': [], 'valid': []}
    for d in dataset_path:
        if not os.path.exists(d) or hasattr(ABSADatasetList, d):
            print('{} dataset is loading from: {}'.format(d, 'https://github.com/yangheng95/ABSADatasets'))
            download_datasets_from_github(os.getcwd())
            search_path = find_dir(os.getcwd(), [d, task], exclude_key=['infer', 'test.'] + filter_key_words, disable_alert=True)
            dataset_file['train'] += find_files(search_path, [d, 'train', task], exclude_key=['.inference', 'test.'] + filter_key_words)
            dataset_file['test'] += find_files(search_path, [d, 'test', task], exclude_key=['inference', 'train.'] + filter_key_words)
            dataset_file['valid'] += find_files(search_path, [d, 'valid', task], exclude_key=['inference', 'train.'] + filter_key_words)
        else:
            dataset_file['train'] = find_files(d, ['train', task], exclude_key=['.inference', 'test.'] + filter_key_words)
            dataset_file['test'] = find_files(d, ['test', task], exclude_key=['.inference', 'train.'] + filter_key_words)
            dataset_file['valid'] = find_files(d, ['valid', task], exclude_key=['.inference', 'train.'] + filter_key_words)

    return dataset_file
